(ns puppetlabs.redrover.postgres
  (:require [cheshire.core :as json]
            [clj-yaml.core :as yaml]
            [clojure.java.io :as io]
            [clojure.java.jdbc.deprecated :as jdbc]
            [clojure.string :as str]
            [puppetlabs.jdbc-util.core :as jdbcu]
            [puppetlabs.redrover.file :refer [read-file]]
            [slingshot.slingshot :refer [try+ throw+]])
  (:import org.postgresql.util.PGobject))

;; necessary to turn the PGobjects created by the row_to_json postgres function
;; into ordinary old clojure maps.
(extend-protocol clojure.java.jdbc/IResultSetReadColumn
  PGobject
  (result-set-read-column [pgobj metadata idx]
    (let [pg-type (.getType pgobj)
          pg-value (.getValue pgobj)]
      (case pg-type
        "json" (json/decode pg-value true)
        pg-value))))

(defn group-export-script*
  []
  (read-file (-> "sql/export-groups.sql" io/resource)))

(defn node-export-script*
  []
  (read-file (-> "sql/export-nodes.sql" io/resource)))

(def group-export-script (memoize group-export-script*))
(def node-export-script (memoize node-export-script*))

(defn serialized-row-object?
  [x]
  (let [f1-f2-map? #(and (map? %) (= (set (keys %)) #{:f1 :f2}))]
    (and (sequential? x)
         (every? f1-f2-map? x))))

(defn unpack-serialized-row-objects
  "The row_to_json function in Postgres serializes the row objects used for the
  variables and classes of groups & nodes as vectors of maps of the form
  `{\"f1\" \"key1\", \"f2\" \"val1\"}`, rather than just
  `{\"key1\" \"val1\", \"key2\" \"val2\"}`. This function recursively transforms
  a vector of maps with the `f1` and `f2` keys into a single map with all the
  keys and values."
  [vec-of-f1-f2-maps]
  (let [deserialized-map (reduce (fn [m {:keys [f1 f2]}] (assoc m f1 f2))
                                 {}
                                 vec-of-f1-f2-maps)]
    (into {} (for [[k v] deserialized-map
                   :let [v' (cond
                              (serialized-row-object? v) (unpack-serialized-row-objects v)
                              (string? v) (yaml/parse-string v)
                              :else v)]]
               [(keyword k) v']))))

(defn unpack-classes-and-variables
  "Given a row map that contains serialized row objects in its `:classes` and
  `:variables` keys, returns a new map with those keys' values transformed into
  regular maps using `unpack-serialized-row-objects`."
  [row-map]
  (-> row-map
    (update-in [:variables] unpack-serialized-row-objects)
    (update-in [:classes] unpack-serialized-row-objects)))

(defn prettify-subname
  [db]
  (-> (:subname db)
    (str/replace #"^//" "")
    (str/replace #"\?.+?$" "")))

(defn with-jdbc-connection
  "Executes the given thunk in the context of a connection to db,
  catching Postgres errors and throwing slingshot errors. When the
  errors are caused by connection/authentication failures, some relevant
  information will be included in the :msg key."
  [db f]
  (try+ (jdbc/with-connection db
          (f))
    (catch org.postgresql.util.PSQLException e
      (let [code (.getSQLState e)]
        (case code
          "08001" (throw+ {:kind ::database-unreachable
                           :msg (str "The database specified by '" (prettify-subname db) "' could "
                                     "not be reached. Please check the database host, port, "
                                     "and name.")})
          "28000" (throw+ {:kind ::authentication-error
                           :msg (str "There was an authentication error while connecting "
                                     "to the database specified by '" (prettify-subname db) "' with "
                                     "the username '" (:user db) "'. Please make sure that the "
                                     "username and password are correct, and that the user has "
                                     "permission to access the database.")})
          (throw+ {:kind ::unknown-error
                   :msg (str "There was an error connecting to the database specified by '"
                             (prettify-subname db) "': " (str e))}))))
    (catch IllegalArgumentException e
      (if (re-find #".*is missing a required parameter.*" (str e))
        ;; JDBC complained about the db-spec
        (throw+ {:kind ::bad-configuration
                 :msg (str "The database was not sufficiently specified. Please make sure "
                           "that a valid hostname and/or database name have been supplied, "
                           "either using --db-host/--db-name or by passing in a valid "
                           "dashboard configuration file with the --database setting.")})
        ;; this is caused by something else and we should re-throw
        (throw e)))))

(defn export-from-dashboard
  [db sql]
  (with-jdbc-connection db
    #(->> (jdbcu/query db [sql])
       (map :row_to_json)
       (map unpack-classes-and-variables))))

(defn get-dashboard-groups
  [db]
  (export-from-dashboard db (group-export-script)))

(defn get-dashboard-nodes
  [db]
  (export-from-dashboard db (node-export-script)))
